@@ -10,14 +10,10 @@ if not sys.warnoptions:
 import matplotlib.pyplot as plt
 import pandas as pd
 import numpy as np
-import seaborn as sns
-
-
+from seaborn import heatmap
 
 
 import lightgbm as lgb
-import catboost as ctb
-import xgboost as xgb
 
 
 from sklearn.preprocessing import LabelEncoder
@@ -27,7 +23,11 @@ from sklearn.metrics import classification_report, confusion_matrix
 from sklearn.metrics import f1_score
 
 
+from itertools import chain, combinations
+from termcolor import colored
 
+from multiprocessing import Pool
+from multiprocessing import cpu_count
 
 
 
@@ -77,6 +77,15 @@ class Snaplib:
         
         
     
+
+
+    class Counter:
+        def __init__(self):
+            self.__counter = 0
+
+        def __call__(self, step=1):
+            self.__counter += step
+            return self.__counter
     
     
         
@@ -406,7 +415,7 @@ class Snaplib:
                             algorithms_list, 
                             X_train, 
                             y_train, 
-                            X_pred, 
+                            X_test, 
                             y_test=None, 
                             task='clsf', 
                             verbose=0
@@ -420,7 +429,7 @@ class Snaplib:
                                         algorithms_list, 
                                         X_train, 
                                         y_train, 
-                                        X_pred, 
+                                        X_test, 
                                         y_test or None, 
                                         task='clsf' or 'regr', 
                                         verbose= 0, 1, 2
@@ -428,7 +437,7 @@ class Snaplib:
         
         algorithms_list = list of algorithms [LGBMClassifier(), XGBClassifier(), CatBoostClassifier()].
         X_train and y_train are data for training list of algorithms.
-        X_pred is dataframe for prediction.
+        X_test is dataframe for prediction.
         y_test optionaly. If exist visualize it as last column on a plot (verbose=True). 
         task='clsf' or 'regr', classification or regression
         verbose = 0 mute, 1 verbose.
@@ -438,7 +447,7 @@ class Snaplib:
             raise TypeError('algorithms_list must be of list type.')
         if not isinstance(X_train, pd.core.frame.DataFrame):
             raise TypeError('The X__train must be a pandas.core.frame.DataFrame instance.')
-        if not isinstance(X_pred, pd.core.frame.DataFrame):
+        if not isinstance(X_test, pd.core.frame.DataFrame):
             raise TypeError('The X__pred must be a pandas.core.frame.DataFrame instance.')
 
         if not isinstance(y_train, pd.core.frame.Series) and not isinstance(y_train, np.ndarray):
@@ -458,14 +467,22 @@ class Snaplib:
 
         stacked_predicts = pd.DataFrame()
         alg_names = []
-        for alg in algorithms_list:
-            alg_name = alg.__class__.__name__[:3]
-            # if y_test is not None and alg_name in ['LGB', 'XGB', 'Cat']:
-            #     model = alg.fit(X_train, y_train, eval_set=[(X_pred, y_test)], early_stopping_rounds=10, verbose=False)
-            # else:
-            model = alg.fit(X_train, y_train)
+        increment = 1
+        for algorithm in algorithms_list:
+            alg = algorithm[0]
+            params = algorithm[1]
+            alg_name = alg.__name__[:3]
+            if alg_name in alg_names:
+                alg_name = alg_name + '_' + str(increment)
+                increment+=1
+            if y_test is not None and alg_name in ['LGB', 'XGB', 'Cat']:
+                model = alg(**params)
+                model.fit(X_train, y_train, eval_set=[(X_test, y_test)], early_stopping_rounds=200, verbose=False)
+            else:
+                model = alg(**params)
+                model.fit(X_train, y_train)
 
-            y_hat = model.predict(X_pred)
+            y_hat = model.predict(X_test)
             if task =='clsf':
                 stacked_predicts[alg_name] = y_hat.astype('int64')
             elif task=='regr':
@@ -481,14 +498,7 @@ class Snaplib:
             if y_test is not None:
                 stacked_predicts['Y_TEST'] = y_test.values
 
-
-
-        y_hat = stacked_predicts.loc[:, 'Y_HAT_STACKED']
-        if verbose:
-            plt.figure(figsize=(5, 10))
-            sns.heatmap(stacked_predicts[-1000:], cbar=False)
-            plt.show()
-        return y_hat
+        return stacked_predicts
 
 
     
@@ -496,18 +506,17 @@ class Snaplib:
     
     
     
+        
     
-    
-
-
 
     def cross_val(self, algorithms, k_fold_dict, metric, task, cv, verbose=0):
-        
+
+                
         ''' 
         Cross Validation method for list of algorithms.
         
         Use case:
-        y_hat = Snaplib().cross_val(algorithms, k_fold_dict, metric, task, cv, verbose=0):
+        score = Snaplib().cross_val(algorithms, k_fold_dict, metric, task, cv, verbose=0):
         
         algorithms_list = list of algorithms like [LGBMClassifier(), XGBClassifier(), CatBoostClassifier()].
         k_fold_dict is a dictionary with the structure:
@@ -519,6 +528,9 @@ class Snaplib:
                         'train_y' : [seri_0, seri_1, seri_2],
                         'test_y'  : [seri_0, seri_1, seri_2],
                        }
+
+        Get k_fold_dict:
+        k_fold_dict = Snaplib().k_folds_split(df, target_name_str, K_FOLD)
               
         metric is a metric like f1_score or mean_absolute_error.
         task='clsf' or 'regr', classification or regression.
@@ -541,48 +553,51 @@ class Snaplib:
             raise TypeError('verbose must be of int type or bool.')
 
 
-        results=[]
-        test_y_all = np.array([])
-        pred_all = np.array([])
-        if task == 'clsf':
-            length = len(k_fold_dict['train_y'][0].value_counts().index)
-            cm_base = np.zeros([length, length])
-            # cm_base = np.array([[0, 0], [0, 0]])
+        global num_combinations
+        global all_prediction_df
+        global counter
+        global cpu_num
+    
+        
+        results=[]    
+        all_prediction_df = pd.DataFrame()
+        counter = self.Counter()
+        research_best_score=dict()
+        cpu_num = cpu_count()
+        
+        
+        def powerset(names_ls):
+            return chain.from_iterable(combinations(names_ls, r) for r in range(len(names_ls)+1))
 
 
+        
+        
         for k in range(0, cv):
-            if len(algorithms) > 1:
-                pred = self.fit_predict_stacked(algorithms, 
-                                                k_fold_dict['train_X'][k], 
-                                                k_fold_dict['train_y'][k], 
-                                                k_fold_dict['test_X'][k], 
-                                                k_fold_dict['test_y'][k], 
-                                                task,
-                                                verbose
-                                            )
-            else:
-                alg = algorithms[0]
-                alg.fit(k_fold_dict['train_X'][k], k_fold_dict['train_y'][k])
-                pred = alg.predict(k_fold_dict['test_X'][k])
-
-            if verbose:
-                if task == 'clsf':
-                    cm = confusion_matrix(k_fold_dict['test_y'][k], pred)
-                    cm_base = cm_base + cm
-
-            test_y_all = np.concatenate(([test_y_all, k_fold_dict['test_y'][k]]), axis=None)
-            pred_all = np.concatenate((pred_all, pred), axis=None)
+            pred_frame = self.fit_predict_stacked(algorithms, 
+                                                    k_fold_dict['train_X'][k], 
+                                                    k_fold_dict['train_y'][k], 
+                                                    k_fold_dict['test_X'][k], 
+                                                    k_fold_dict['test_y'][k], 
+                                                    task,
+                                                    verbose,
+                                                    )
 
+            pred = pred_frame.loc[:, 'Y_HAT_STACKED']
             score = metric(k_fold_dict['test_y'][k], pred)
             results.append(score)
 
+            if verbose:
+                all_prediction_df = pd.concat([all_prediction_df, pred_frame], ignore_index=True)
+
         results_std = np.std(results)
         score = sum(results) / len(results)
 
+
         if verbose:
+            all_prediction_df.reset_index(drop=True, inplace=True)
             for alg in algorithms:
-                print(alg.__class__.__name__)
-                
+                print(alg[0].__name__)
+
             print('')
             if str(metric).split('.')[0] == 'functools':
                 metric_name = str(metric).split('function ')[1].split(' ')[0]
@@ -592,12 +607,140 @@ class Snaplib:
             print('\n', results, '\n')
 
             if task == 'clsf':
-                print('\n', classification_report(test_y_all, pred_all), '\n')
+                print('\n', classification_report(all_prediction_df['Y_TEST'], all_prediction_df.loc[:, 'Y_HAT_STACKED']), '\n')
                 plt.figure(figsize=(3, 3))
-                sns.heatmap(cm_base, annot=True, cmap="Blues", fmt='.0f',  cbar=False)
+                cm = confusion_matrix(all_prediction_df.loc[:, 'Y_TEST'], all_prediction_df.loc[:, 'Y_HAT_STACKED'])
+                heatmap(cm, annot=True, cmap="Blues", fmt='.0f',  cbar=False)
                 plt.show()
 
-        return score
+
+            rng = np.random.default_rng()
+            idx = set(rng.integers(len(all_prediction_df)-1, size=200))
+
+            data = all_prediction_df.loc[idx,:]
+            data.sort_values(by='Y_TEST', ascending=True, inplace=True)
+            data.reset_index(drop=True, inplace=True)
+            if task =='clsf':
+                plt.figure(figsize=(5, 10))
+                plt.pcolor(data, cmap='Blues_r')
+                cols = list(data.columns)
+                plt.xticks(np.arange(0.5, len(cols), 1), cols, rotation=80)
+                plt.show()
+
+            elif task=='regr':
+                plt.figure(figsize=(20, 10))
+                for col in data.columns:
+                    width = 5 if col == 'Y_TEST' else 1
+                    plt.plot(data[col], label=col, linewidth=width)
+                plt.legend()
+                plt.xticks([])
+                plt.title('RANDOM 200 cases sorted by test value')
+                plt.show()
+
+                
+            # BEST COMBINATION RESEARCH
+            all_prediction_df_columns = list(all_prediction_df.columns)
+            if len(all_prediction_df_columns) > 3:
+                print('\nThe Best Algorithms Combination:')
+                alg_names = all_prediction_df_columns[:]
+                alg_names.remove('Y_TEST')
+                alg_names.remove('Y_HAT_STACKED')
+
+                all_combinations = list(powerset(alg_names))
+                num_combinations = len(all_combinations) -1
+                
+                results_pool = []
+                if task =='clsf':
+                    with Pool(cpu_num) as p:
+                        results_pool.append(p.map(self.get_score_clsf, all_combinations[1:]))
+                elif task=='regr':
+                    with Pool(cpu_num) as p:
+                        results_pool.append(p.map(self.get_score_regr, all_combinations[1:]))
+
+                for line in results_pool[0]:
+                    key, value = line[0], line[1]
+                    research_best_score[key] = value
+                    
+                print(f'\r{num_combinations} / {num_combinations} processed', end='', flush=True)
+                
+
+                best_list=[]
+                ascending = True if task=='clsf' else False
+                research_best_scoredict = sorted(research_best_score.items(), reverse=ascending)
+                print(f'\n{metric_name}:          combination\n')
+                i=0
+                color='grey'
+                for k, v in research_best_scoredict:
+                    ks, vs  = str(k), str(v)
+                    if i == 0:
+                        color='green'
+                        best_list = v[:]
+                    if vs == str(alg_names):
+                        color='blue'
+                        vs = vs + ' - full stack of algorithms'
+                    print("{:30s} {:100s} ".format(colored(ks, color), colored(vs, color)))
+                    color='grey'
+                    i+=1
+
+                    
+                # THE  BEST  SHOW  PLOTS
+                if task =='clsf':
+                    best_df = data[list(best_list)]
+                    best_df['Y_HAT_STACKED'] = data.loc[:, list(best_list)].mode(axis=1)[0].astype('int64')
+                    best_df['Y_TEST'] = data['Y_TEST']
+
+                    f, ax = plt.subplots(1, 2, figsize=(10, 10))
+
+                    plt.subplot(1, 2, 1)
+                    plt.pcolor(data, cmap='Blues_r')
+                    cols = list(all_prediction_df_columns)
+                    plt.xticks(np.arange(0.5, len(cols), 1), cols, rotation=80)
+                    ax[0].set_title('ALL ALGORITHMS')
+
+                    plt.subplot(1, 2, 2)
+                    plt.pcolor(best_df, cmap='Blues_r')
+                    cols = list(best_list)+['Y_HAT_STACKED']+['Y_TEST']
+                    plt.xticks(np.arange(0.5, len(cols), 1), cols, rotation=80)
+                    ax[1].set_title('BEST COMPOSITION')
+
+                    f.tight_layout()
+                    plt.show()
+
+                elif task=='regr':
+                    plt.figure(figsize=(20, 10))
+                    y_hat_all = data.loc[:, list(alg_names)].mean(axis=1)
+                    y_hat_best = data.loc[:, list(best_list)].mean(axis=1)
+
+                    plt.plot(y_hat_best, label=str(best_list), linewidth=1, color='green')
+                    plt.plot(y_hat_all, label=str(alg_names), linewidth=1, color='blue')
+                    plt.plot(data['Y_TEST'], label='Y_TEST', linewidth=5, color='red')
+                    plt.legend()
+                    plt.xticks([])
+                    plt.title('Compare all algorithms and the best')
+                    plt.show()
+                    
+                del num_combinations
+                del all_prediction_df
+                del counter
+                del cpu_num
+
+            return score
+        
+    @staticmethod
+    def get_score_clsf(alg_set):
+        print(f'\r{counter(cpu_num)} / {num_combinations} processed', end='', flush=True)
+        algs_list = list(alg_set)
+        y_hat_set = all_prediction_df.loc[:, algs_list].mode(axis=1)[0].astype('int64')
+        score_set = metric(all_prediction_df.loc[:, ['Y_TEST']], y_hat_set)
+        return score_set, algs_list
+    
+    @staticmethod
+    def get_score_regr(alg_set):
+        print(f'\r{counter(cpu_num)} / {num_combinations} processed', end='', flush=True)
+        algs_list = list(alg_set)
+        y_hat_set = all_prediction_df.loc[:, algs_list].mean(axis=1)
+        score_set = metric(all_prediction_df.loc[:, ['Y_TEST']], y_hat_set)
+        return score_set, algs_list
     
     
 
@@ -713,6 +856,8 @@ class Snaplib:
         alg_names = []
         for alg in algorithms_list:
             alg_name = alg.__class__.__name__[:3]
+            if alg_name in alg_names:
+                alg_name = alg_name + '_another'
             y_hat = alg.predict(X_pred)
             if task =='clsf':
                 stacked_predicts[alg_name] = y_hat.astype('int64')
@@ -862,6 +1007,8 @@ class Snaplib:
         
         for alg in algorithms_list:
             filename = alg.__class__.__name__ + '.sav'
+            if filename in names:
+                filename =  alg.__class__.__name__  + '_another' + '.sav'
             path = directory + filename
             pickle.dump(alg, open(path, 'wb'))
             names.append(filename)
@@ -1136,6 +1283,7 @@ class Snaplib:
 
     def recover_data(self,
                      df_0, 
+                     device='cpu',
                      verbose = 1,
                      ):
         ''' 
@@ -1143,20 +1291,22 @@ class Snaplib:
 
         Use case:
         df = Snaplib().recover_data(df, verbose=True, stacking=True)
-
+        device must be "cpu" or "gpu". Sometime small datasets work faster with cpu.
         if set verbose = if True algorithm runs cross validation tests and print results of tests for decision making.
         And ensemble decrise train/test leakage.
         '''
 
         if not isinstance(df_0, pd.core.frame.DataFrame):
             raise TypeError('The df must be a pandas.core.frame.DataFrame instance.')
+        if device !='cpu' and device != 'gpu':
+            raise ValueError('device must be "cpu" or "gpu".')
         if type(verbose) != int and type(verbose) != bool:
             raise TypeError('verbose must be of int type or bool.')
 
 
         counter_predicted_values = 0
         CLASS_VALUE_COUNTS = 30
-        K_FOLDS = 3
+        K_FOLDS = 4
         miss_indeces = None
 
 
@@ -1239,7 +1389,6 @@ class Snaplib:
 
 
             target_values_counted = y.value_counts()
-            last_item = target_values_counted.tail(1).item()
             len_target_values_counted = len(target_values_counted)
 
 
@@ -1269,12 +1418,11 @@ class Snaplib:
                 # Test
                 if verbose:
                     print('CLASSIFIER cross validation:')                    
-                    results=[]
                     test_y_all = np.array([])
                     pred_all = np.array([])
 
                     for k in range(0, K_FOLDS):
-                        lgb_class = lgb.LGBMClassifier(random_state=0, n_jobs=-1)
+                        lgb_class = lgb.LGBMClassifier(random_state=0, n_jobs=-1, device=device)
                         lgb_class.fit(k_fold_dict['train_X'][k], k_fold_dict['train_y'][k])
                         pred = lgb_class.predict(k_fold_dict['test_X'][k])
                         test_y_all = np.concatenate(([test_y_all, k_fold_dict['test_y'][k]]), axis=None)
@@ -1297,7 +1445,7 @@ class Snaplib:
                     
 
                 # Final prediction
-                lgb_class = lgb.LGBMClassifier(random_state=0, n_jobs=-1)
+                lgb_class = lgb.LGBMClassifier(random_state=0, n_jobs=-1, device=device)
                 lgb_class.fit(X, y)
                 pred_miss = lgb_class.predict(miss_df)
 
@@ -1312,14 +1460,13 @@ class Snaplib:
                 if verbose:
                     print('REGRESSOR cross validation:')
 
-                    results=[]
                     test_y_all = np.array([])
                     pred_all = np.array([])
                     for k in range(0, K_FOLDS):
-                        lgb_reg = lgb.LGBMRegressor(n_jobs=-1, random_state=0)
+                        lgb_reg = lgb.LGBMRegressor(n_jobs=-1, random_state=0, device=device)
                         lgb_reg.fit(k_fold_dict['train_X'][k], k_fold_dict['train_y'][k])
                         pred = lgb_reg.predict(k_fold_dict['test_X'][k])
-                        if y.min() == 0.0:
+                        if y.min() >= 0.0:
                             pred[pred < 0] = 0
                         test_y_all = np.concatenate(([test_y_all, k_fold_dict['test_y'][k]]), axis=None)
                         pred_all = np.concatenate((pred_all, pred), axis=None)
@@ -1340,10 +1487,10 @@ class Snaplib:
                     print(f'first 10 y_pred: {list(np.round(pred, 1))}\n')
 
                 # Final prediction
-                lgb_reg = lgb.LGBMRegressor(random_state=0, n_jobs=-1)
+                lgb_reg = lgb.LGBMRegressor(random_state=0, n_jobs=-1, device=device)
                 lgb_reg.fit(X, y)
                 pred_miss = lgb_reg.predict(miss_df)
-                if y.min() == 0:
+                if y.min() >= 0:
                     pred_miss[pred_miss < 0] = 0
 
                 pred_miss = np.expm1(pred_miss)
